package com.tibco.accelerators.operators.patterns;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;

import com.netrics.likeit.NetricsConMgr;
import com.netrics.likeit.NetricsEmbedded;
import com.netrics.likeit.NetricsEmbeddedSettings;
import com.netrics.likeit.NetricsException;
import com.netrics.likeit.NetricsQuery;
import com.netrics.likeit.NetricsRecord;
import com.netrics.likeit.NetricsSearchCfg;
import com.netrics.likeit.NetricsSearchOpts;
import com.netrics.likeit.NetricsSearchResponse;
import com.netrics.likeit.NetricsSearchResult;
import com.netrics.likeit.NetricsServerInterface;
import com.netrics.likeit.NetricsTable;
import com.streambase.sb.*;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.operator.*;

public class PatternsOperator extends Operator implements Parameterizable {

	public static final long serialVersionUID = 1511355949887L;
	private Logger logger;
	
	// Properties
	private String connectionType;
	private String consoleLogFilename;
	private int maxTableMemory;
	private String embeddedLibPath;
	private String hostMachine;
	private int hostPort;
	private String connectionPooling;
	private boolean statusPort;

	private Schema tableSchema;
	private String tableName;
	private boolean hasAttributes;

	private double useAbsoluteCutoff;
	
	private String displayName = "TIBCO Patterns Operator";
	
	// Local variables
	private int inputPorts = 1;
	private int outputPorts = 2;
	private Schema[] outputSchemas; // caches the Schemas given during init() for use at processTuple()

	// Local References
	private NetricsServerInterface nsi;
	
	/**
	* The constructor is called when the Operator instance is created, but before the Operator 
	* is connected to the StreamBase application. We recommended that you set the initial input
	* port and output port count in the constructor by calling setPortHints(inPortCount, outPortCount).
	* The default is 1 input port, 1 output port. The constructor may also set default values for 
	* operator parameters. These values will be displayed in StreamBase Studio when a new instance
	* of this operator is  dragged to the canvas, and serve as the default values for omitted
	* optional parameters.
	 */
	public PatternsOperator() {
		super();
		logger = getLogger();
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setConsoleLogFilename("");
		setMaxTableMemory(100000);
		setEmbeddedLibPath("");
		setHostMachine("localhost");
		setHostPort(5051);
		setStatusPort(true);
		setPortHints(inputPorts, outputPorts);

		setTableName("");
		setHasAttributes(true);
		setTableSchema(null);

		setUseAbsoluteCutoff(0.25D);
	}

	/**
	* The typecheck method is called after the Operator instance is connected in the StreamBase
	* application, allowing the Operator to validate its properties. The Operator class may 
	* change the number of input or output ports by calling the requireInputPortCount(portCount)
	* method or the setOutputSchema(schema, portNum) method. If the verifyInputPortCount method 
	* is passed a different number of ports than the Operator currently has, a PortMismatchException
	* (subtype of TypecheckException) is thrown.
	*/
	public void typecheck() throws TypecheckException {
		// typecheck: require a specific number of input ports
		requireInputPortCount(inputPorts);

		// Validate configuration
		if(this.getConnectionType().equals("Embedded") && (this.getEmbeddedLibPath() == null || this.getEmbeddedLibPath().length() == 0))
			throw new TypecheckException("Embedded lib path not specified");
		
		// Validate input schema
		Schema inputSchema = getInputSchema(0);

		if(!inputSchema.hasField("record") && !inputSchema.hasField("search"))
			throw new TypecheckException("Input schema must define record or search");

		// Validate record
		if(inputSchema.hasField("record")) {
			try {
				Field recordField = inputSchema.getField("record");
				if(recordField.getDataType() == DataType.TUPLE) {
					Schema recordSchema = recordField.getSchema();
					if(!recordSchema.hasField("key"))
						throw new TypecheckException("Record must define key");
					if(!recordSchema.hasField("add") && !recordSchema.hasField("update") && !recordSchema.hasField("delete"))
						throw new TypecheckException("Record must define add, update, or delete");
				}
				else
					throw new TypecheckException("Record must be tuple type");
			} catch (TupleException e) {
				throw new TypecheckException(e);
			}
		}
		
		// Validate search
		if(inputSchema.hasField("search")) {
			try {
				Field searchField = inputSchema.getField("search");
				if(searchField.getDataType() == DataType.TUPLE) {
					Schema searchSchema = searchField.getSchema();
					if(!searchSchema.hasField("term"))
						throw new TypecheckException("Search must define term");
				}
				else
					throw new TypecheckException("Search must be tuple type");
			} catch (TupleException e) {
				throw new TypecheckException(e);
			}
		}
		
		
		
		// Setup output schemas
		List<Field> operationFields = new ArrayList<Field>();
		for(int idx = 0; idx < inputSchema.getFieldCount(); idx++) {
			Field f = inputSchema.getField(idx);
			if(!f.getName().equals("record") && !f.getName().equals("search"))
				operationFields.add(f);
		}
		
		// If there is a search input, create a search response schema
		if(inputSchema.hasField("search")) {
			Schema attributesSchema = new Schema(null, Schema.createField(DataType.STRING, "name"),
					Schema.createField(DataType.STRING, "value"));
			List<Field> searchResultFields = new ArrayList<Field>();
			searchResultFields.add(Schema.createField(DataType.STRING, "key"));
			searchResultFields.add(Schema.createField(DataType.DOUBLE, "score"));
			searchResultFields.add(Schema.createField(DataType.INT, "rank"));
			if(this.getTableSchema() != null) {
				for(int idx = 0; idx < this.getTableSchema().getFieldCount(); idx++)
					searchResultFields.add(Schema.createField(DataType.STRING, this.getTableSchema().getField(idx).getName()));
			}
			if(this.getHasAttributes() == true) {
				searchResultFields.add(Schema.createListField("attributes", CompleteDataType.forTuple(attributesSchema)));				
			}
			Schema searchResultSchema = new Schema(null, searchResultFields);
			operationFields.add(Schema.createTupleField("search", new Schema(null, 
					Schema.createField(DataType.STRING, "term"),
					Schema.createListField("match", CompleteDataType.forTuple(searchResultSchema)))));
		}
		
		// Set output schema
		Schema outputOperationSchema = new Schema(null, operationFields);
		setOutputSchema(0, outputOperationSchema);

		// If status port enabled, set output schema
		if(this.getStatusPort() == true) {
			Schema outputStatusSchema = new Schema(null,
					Schema.createField(DataType.STRING, "operation"),
					Schema.createField(DataType.STRING, "type"),
					Schema.createField(DataType.STRING, "result"));
			setOutputSchema(1, outputStatusSchema);
		}
	}

	/**
	* This method will be called by the StreamBase server for each Tuple given
	* to the Operator to process. This is the only time an operator should 
	* enqueue output Tuples.
	* @param inputPort the input port that the tuple is from (ports are zero based)
	* @param tuple the tuple from the given input port
	* @throws StreamBaseException Terminates the application.
	*/
	public void processTuple(int inputPort, Tuple tuple) throws StreamBaseException {		
		try {
			if(tuple.isNull("record") == false)
				this.processRecordTuple(tuple);
			else if(tuple.isNull("search") == false)
				this.processSearchTuple(tuple);
		}
		catch(StreamBaseException e) {
			e.printStackTrace();
			throw e;
		}
	}

	/**
	 * Process a record tuple.
	 * @param tuple
	 * @throws StreamBaseException
	 */
	private void processRecordTuple(Tuple tuple) throws StreamBaseException {
		Tuple recordTuple = tuple.getTuple("record");

		// Get the key
		String key = recordTuple.getString("key");
		
		// See if add, upd, or del
		if(recordTuple.isNull("add") == false) {
			try {
				// Extract the add and create the record
				Tuple recordAddTuple = recordTuple.getTuple("add");
				NetricsRecord record = this.createRecordFromTuple(recordAddTuple, key);

				// Add the record to the table on the Patterns server
				this.nsi.recadd(this.getTableName(), record);
				
				// Send status if enabled
				if(this.getStatusPort() == true) {
					Tuple out = this.outputSchemas[1].createTuple();
					out.setField("operation", "record");
					out.setField("type", "add");
					out.setField("result", "success");
					sendOutput(1, out);
				}
			} catch (NetricsException | IOException e) {
				if(this.getStatusPort() == true) {
					Tuple out = this.outputSchemas[1].createTuple();
					out.setField("operation", "record");
					out.setField("type", "add");
					out.setField("result", e);
					sendOutput(1, out);
				}
				logger.error(e.getMessage());
				//e.printStackTrace();
			} 
			
		}
		else if(recordTuple.isNull("update") == false) {
			try {
				// Extract the update and create the record
				Tuple recordUpdateTuple = recordTuple.getTuple("update");
				NetricsRecord record = this.createRecordFromTuple(recordUpdateTuple, key);
				
				// Update the record to the table on the Patterns server
				this.nsi.recupdate(this.getTableName(), record);
				
				// Send status if enabled
				if(this.getStatusPort() == true) {
					Tuple out = this.outputSchemas[1].createTuple();
					out.setField("operation", "record");
					out.setField("type", "update");
					out.setField("result", "success");
					sendOutput(1, out);
				}
			} catch (NetricsException | IOException e) {
				if(this.getStatusPort() == true) {
					Tuple out = this.outputSchemas[1].createTuple();
					out.setField("operation", "record");
					out.setField("type", "update");
					out.setField("result", e);
					sendOutput(1, out);
				}
				e.printStackTrace();
			} 
		}
		else if(recordTuple.isNull("delete") == false) {
			try {
				// Delete this record from the table on the Patterns server
				this.nsi.recdelete(this.getTableName(), key);
				
				// Send status if enabled
				if(this.getStatusPort() == true) {
					Tuple out = this.outputSchemas[1].createTuple();
					out.setField("operation", "record");
					out.setField("type", "delete");
					out.setField("result", "success");
					sendOutput(1, out);
				}				
			} catch (IOException | NetricsException e) {
				if(this.getStatusPort() == true) {
					Tuple out = this.outputSchemas[1].createTuple();
					out.setField("operation", "record");
					out.setField("type", "delete");
					out.setField("result", e);
					sendOutput(1, out);
				}
				e.printStackTrace();
			}
		}
		
		// Send the tuple output
		Tuple out = this.getTupleOut(tuple);
		sendOutput(0, out);
		
		// DEBUG Write contents of table
		/*try {
			com.netrics.likeit.NetricsTableCursor cursor = new com.netrics.likeit.NetricsTableCursor(this.getTableName(), 1000);
			NetricsRecord[] records = this.nsi.recnext(cursor);
			System.out.println(">>>>>" + this.getTableName() + "<<<<<");
			for(int idx = 0; idx < records.length; idx++) {
				System.out.println(records[idx]);
			}
		} catch (IOException | NetricsException e) {
			e.printStackTrace();
		}*/
	}
	
	/**
	 * Creates a new Patterns record from the specified data tuple. 
	 * @param tuple
	 * @param key
	 * @return
	 * @throws NullValueException
	 * @throws TupleException
	 */
	private NetricsRecord createRecordFromTuple(Tuple tuple, String key) throws NullValueException, TupleException {
		// Build the field list
		List<String> fields = new ArrayList<String>();
		for(int idx = 0; idx < this.getTableSchema().getFieldCount(); idx++) {
			Field f = this.getTableSchema().getField(idx);
			fields.add(tuple.getField(f.getName()).toString());
		}
		// Build the attributes list
		List<?> attributesList = null;
		if(this.getHasAttributes() == true) {
			attributesList = tuple.getList("attributes");
		}

		// Create the record
		NetricsRecord record;
		if(attributesList == null) {
			record = new NetricsRecord(key, fields.toArray(new String[fields.size()]));
		}
		else {
			String[] attributeNames = new String[attributesList.size()];
			String[] attributeValues = new String[attributesList.size()];
			for(int idx = 0; idx < attributesList.size(); idx++) {
				Tuple attributeTuple = (Tuple) attributesList.get(idx);
				attributeNames[idx] = attributeTuple.getString("name");
				if(attributeTuple.isNull("value"))
					attributeValues[idx] = "";
				else
					attributeValues[idx] = attributeTuple.getField("value").toString();
			}
			record = new NetricsRecord(key, fields.toArray(new String[fields.size()]), attributeNames, attributeValues);					
		}	
		
		return record;
	}
	
	/**
	 * Process a search tuple.
	 * @param tuple
	 * @throws NullValueException
	 * @throws TupleException
	 */
	private void processSearchTuple(Tuple tuple) throws NullValueException, TupleException {
		Tuple searchTuple = tuple.getTuple("search");

		// Get the table name and term
		String term = searchTuple.getString("term");
		
		NetricsSearchCfg searchCfg = new NetricsSearchCfg(this.getTableName());
		NetricsQuery query = NetricsQuery.Simple(term, null, null);
		searchCfg.setNetricsQuery(query);
		NetricsSearchOpts searchOpts = new NetricsSearchOpts();
		searchOpts.useAbsoluteCutoff(this.getUseAbsoluteCutoff());
		
		try {
			NetricsSearchResponse response = this.nsi.search(searchCfg,  searchOpts);
			NetricsSearchResult[] results = response.getSearchResults();

			// Get response tuple
			Tuple out = this.getTupleOut(tuple);
			
			// Create search out
			Schema searchOutSchema = this.outputSchemas[0].getField("search").getSchema();
			Tuple searchOut = searchOutSchema.createTuple();
			
			Schema matchOutSchema = searchOutSchema.getField("match").getCompleteDataType().getElementType().getSchema();
			Schema attributeOutSchema = null;
			if(this.getHasAttributes() == true) {
				attributeOutSchema = matchOutSchema.getField("attributes").getCompleteDataType().getElementType().getSchema();
			}				

			List<Tuple> matchList = new ArrayList<Tuple>();
			for(int resultIdx = 0; resultIdx < results.length; resultIdx++) {
				NetricsSearchResult result = results[resultIdx];
				Tuple match = matchOutSchema.createTuple();
				match.setString("key", result.getKey());
				match.setDouble("score", result.getMatchScore());
				match.setInt("rank", result.getMatchRank());

				for(int idx = 0; idx < result.getFieldNames().length; idx++) {
					if(!result.getFieldNames()[idx].equals("attributes"))
						match.setString(result.getFieldNames()[idx], result.getField(idx));
				}
				if(this.getHasAttributes() == true) {
					List<Tuple> attributeList = new ArrayList<Tuple>();
					for(int idx = 0; idx < result.getAttrNames().length; idx++) {
						Tuple attribute = attributeOutSchema.createTuple();
						attribute.setString("name", result.getAttrNames()[idx]);
						attribute.setString("value", result.getAttrValues()[idx]);
						attributeList.add(attribute);
					}
					match.setList("attributes", attributeList);
				}
				matchList.add(match);
			}	
			searchOut.setString("term", term);
			searchOut.setList("match", matchList);
			out.setField("search", searchOut);

			// Send output
			sendOutput(0, out);
		} catch (Exception ex) {
			ex.printStackTrace();
		}		
	}
	
	/**
	 * Sends tuple response from the specified input tuple.
	 * @param tuple
	 * @throws StreamBaseException
	 */
	private Tuple getTupleOut(Tuple tuple) throws StreamBaseException {
		// Send the tuple output
		Schema inputSchema = tuple.getSchema();
		Schema outputSchema = outputSchemas[0];
		Tuple out = outputSchema.createTuple();
		for(int idx = 0; idx < inputSchema.getFieldCount(); idx++) {
			Field f = inputSchema.getField(idx);
			if(!f.getName().equals("record") && !f.getName().equals("search"))
				out.setField(f.getName(), tuple.getField(f));
		}
		return out;
	}
	
	/**
	 * If typecheck succeeds, the init method is called before the StreamBase application
	 * is started. Note that your Operator class is not required to define the init method,
	 * unless (for example) you need to perform initialization of a resource such as a JDBC
	 * pool, if your operator is making JDBC calls. StreamBase Studio does not call this
	 * during authoring.
	 */
	public void init() throws StreamBaseException {
		super.init();

		if(this.getConnectionType().equals("Embedded")) {
			NetricsEmbeddedSettings embeddedSettings = new NetricsEmbeddedSettings() ;
			if(this.getConsoleLogFilename() != null && this.getConsoleLogFilename().length() > 0)
				embeddedSettings.consoleLogFilename = this.consoleLogFilename;
			embeddedSettings.maxTableMemory = this.maxTableMemory;

			File embeddedLib = new File(this.embeddedLibPath) ;

			// start the embedded server.
			try {
			     NetricsEmbedded.initialize(embeddedSettings, embeddedLib) ;
				 NetricsConMgr connManager = new NetricsConMgr() ;
				 connManager.setEmbeddedMode(true) ;
				 this.nsi = connManager.newConnection() ;
				 logger.info("Embedded TIBCO Patterns server connected");
			 } catch (NetricsException | IOException ex) {
				 ex.printStackTrace();
			 } 
		}
		else if(this.getConnectionType().equals("Server")) {
			try {
				NetricsConMgr connManager = new NetricsConMgr();
				connManager.setHost(this.getHostMachine());
				connManager.setPort(this.getHostPort());
				if(this.getConnectionPooling().equals("common"))
					connManager.setConnectionPoolingCommon();
				else if(this.getConnectionPooling().equals("local"))
					connManager.setConnectionPoolingLocal();
				else if(this.getConnectionPooling().equals("none"))
					connManager.setConnectionPoolingOff();
				this.nsi = connManager.newConnection();
				logger.info("Connected to TIBCO Patterns server");
			} catch (NetricsException | IOException ex) {
				ex.printStackTrace();
			} 
		}
		
		try {
			List<String> fieldNames = new ArrayList<>();
			for(int idx = 0; idx < this.getTableSchema().getFieldCount(); idx++) {
				Field f = this.getTableSchema().getField(idx);
				fieldNames.add(f.getName());
			}

			if(this.getHasAttributes() == true) {
				fieldNames.add("attributes");
			}

			// Build the field names array
			String[] fieldNamesStr = fieldNames.toArray(new String[fieldNames.size()]);

			// Define the table
			NetricsTable table = new NetricsTable(this.getTableName(), fieldNamesStr, null);

			// Load the table into Patterns 
			this.nsi.tblload(table);
			logger.info("Created table " + this.getTableName());
		} catch (IOException | NetricsException e) {
			e.printStackTrace();
		}
		
		
		// for best performance, consider caching input or output Schema.Field objects for
		// use later in processTuple()
		outputSchemas = new Schema[outputPorts];

		for (int i = 0; i < outputPorts; ++i) {
			outputSchemas[i] = getRuntimeOutputSchema(i);
		}
	}

	/**
	*  The shutdown method is called when the StreamBase server is in the process of shutting down.
	*/
	public void shutdown() {

	}

	@Override
	public URL getIconResource(IconKind iconType) {
		switch (iconType) {
		case CANVAS_OVERLAY_ICON:
			return getClass().getResource("icons/patterns_48.png");
		case PALETTE_ICON_LARGE:
			return getClass().getResource("icons/patterns_32.png");
		case PALETTE_ICON_SMALL:
			return getClass().getResource("icons/patterns_16.png");
		}
		return null;
	}
	
	/***************************************************************************************
	 * The getter and setter methods provided by the Parameterizable object.               *
	 * StreamBase Studio uses them to determine the name and type of each property         *
	 * and obviously, to set and get the property values.                                  *
	 ***************************************************************************************/

	public void setConnectionType(String connectionType) {
		this.connectionType = connectionType;
	}

	public String getConnectionType() {
		return this.connectionType;
	}

	public void setConsoleLogFilename(String consoleLogFilename) {
		this.consoleLogFilename = consoleLogFilename;
	}

	public String getConsoleLogFilename() {
		return this.consoleLogFilename;
	}

	public void setMaxTableMemory(int maxTableMemory) {
		this.maxTableMemory = maxTableMemory;
	}

	public int getMaxTableMemory() {
		return this.maxTableMemory;
	}

	public void setEmbeddedLibPath(String embeddedLibPath) {
		this.embeddedLibPath = embeddedLibPath;
	}

	public String getEmbeddedLibPath() {
		return this.embeddedLibPath;
	}

	public void setHostMachine(String hostMachine) {
		this.hostMachine = hostMachine;
	}

	public String getHostMachine() {
		return this.hostMachine;
	}

	public void setHostPort(int hostPort) {
		this.hostPort = hostPort;
	}

	public int getHostPort() {
		return this.hostPort;
	}

	public void setConnectionPooling(String connectionPooling) {
		this.connectionPooling = connectionPooling;
	}

	public String getConnectionPooling() {
		return this.connectionPooling;
	}

	public void setStatusPort(boolean statusPort) {
		this.outputPorts = statusPort == true ? 2 : 1;
		setPortHints(inputPorts, outputPorts);
		this.statusPort = statusPort;
	}

	public boolean getStatusPort() {
		return this.statusPort;
	}

	
	// Table
	public void setTableName(String tableName) {
		this.tableName = tableName;
	}

	public String getTableName() {
		return this.tableName;
	}

	public void setHasAttributes(boolean hasAttributes) {
		this.hasAttributes = hasAttributes;
	}

	public boolean getHasAttributes() {
		return this.hasAttributes;
	}

	public void setTableSchema(Schema tableSchema) {
		this.tableSchema = tableSchema;
	}

	public Schema getTableSchema() {
		return this.tableSchema;
	}

	
	// Search
	public void setUseAbsoluteCutoff(double useAbsoluteCutoff) {
		this.useAbsoluteCutoff = useAbsoluteCutoff;
	}

	public double getUseAbsoluteCutoff() {
		return this.useAbsoluteCutoff;
	}

	
	/** For detailed information about shouldEnable methods, see interface Parameterizable java doc 
	 *  @see Parameterizable 
	 */

	public boolean shouldEnableConnectionType() {
		return true;
	}

	public boolean shouldEnableConsoleLogFilename() {
		if(this.connectionType.equals("Embedded"))
			return true;
		return false;
	}

	public boolean shouldEnableMaxTableMemory() {
		if(this.connectionType.equals("Embedded"))
			return true;
		return false;
	}

	public boolean shouldEnableEmbeddedLibPath() {
		if(this.connectionType.equals("Embedded"))
			return true;
		return false;
	}

	public boolean shouldEnableHostMachine() {
		if(this.connectionType.equals("Server"))
			return true;
		return false;
	}

	public boolean shouldEnableHostPort() {
		if(this.connectionType.equals("Server"))
			return true;
		return false;
	}

	public boolean shouldEnableConnectionPooling() {
		if(this.connectionType.equals("Server"))
			return true;
		return false;
	}

	public boolean shouldEnableStatusPort() {
		return true;
	}

	// Table
	public boolean shouldEnableTableSchema() {
		return true;
	}

	public boolean shouldEnableTableName() {
		return true;
	}

	public boolean shouldEnableHasAttributes() {
		return true;
	}
	
	
	// Search
	public boolean shouldEnableUseAbsoluteCutoff() {
		return true;
	}

}
